'''
Ahocorasick implementation entirely written in python.
Supports unicode.

Quite optimized, the code may not be as beautiful as you like,
since inlining and so on was necessary

This library is optimized for the cPython interpreter.
I will most likely run slower with pypy, etc.

Created on Jan 5, 2016

@author: Frederik Petersen (fp@abusix.com)
'''

from builtins import object


class State(object):
    __slots__ = ['identifier', 'symbol', 'success', 'transitions', 'parent',
                 'matched_keyword', 'longest_strict_suffix']

    def __init__(self, identifier, symbol=None,  parent=None, success=False):
        self.symbol = symbol
        self.identifier = identifier
        self.transitions = {}
        self.parent = parent
        self.success = success
        self.matched_keyword = None
        self.longest_strict_suffix = None


class KeywordTree(object):

    def __init__(self, case_insensitive=False):
        '''
        @param case_insensitive: If true, case will be ignored when searching.
                                 Setting this to true will have a positive
                                 impact on performance.
                                 Defaults to false.
        @param over_allocation: Determines how big initial transition arrays
                                are and how much space is allocated in addition
                                to what is essential when array needs to be
                                resized. Default value 2 seemed to be sweet
                                spot for memory as well as cpu.
        '''
        self._zero_state = State(0)
        self._counter = 1
        self._finalized = False
        self._case_insensitive = case_insensitive

    def add(self, keyword):
        '''
        Add a keyword to the tree.
        Can only be used before finalize() has been called.
        Keyword should be str or unicode.
        '''
        if self._finalized:
            raise ValueError('KeywordTree has been finalized.' +
                             ' No more keyword additions allowed')
        original_keyword = keyword
        if self._case_insensitive:
            keyword = keyword.lower()
        if len(keyword) <= 0:
            return
        current_state = self._zero_state
        for char in keyword:
            try:
                current_state = current_state.transitions[char]
            except KeyError:
                next_state = State(self._counter, parent=current_state,
                                   symbol=char)
                self._counter += 1
                current_state.transitions[char] = next_state
                current_state = next_state
        current_state.success = True
        current_state.matched_keyword = original_keyword

    def search(self, text):
        '''
        Search a text for any occurence of a keyword.
        Returns when one keyword has been found.
        Can only be called after finalized() has been called.
        O(n) with n = len(text)
        @return: 2-Tuple with keyword and startindex in text.
                 Or None if no keyword was found in the text.
        '''
        if not self._finalized:
            raise ValueError('KeywordTree has not been finalized.' +
                             ' No search allowed. Call finalize() first.')
        if self._case_insensitive:
            text = text.lower()
        current_state = self._zero_state
        for idx, symbol in enumerate(text):
            current_state = current_state.transitions.get(symbol,
                                                          self._zero_state)

            if current_state.success:
                keyword = current_state.matched_keyword
                return (keyword, idx + 1 - len(keyword))

    def finalize(self):
        '''
        Needs to be called after all keywords have been added and
        before any searching is performed.
        '''
        if self._finalized:
            raise ValueError('KeywordTree has already been finalized.')
        self._zero_state.longest_strict_suffix = self._zero_state
        self.search_lss_for_children(self._zero_state)
        self._finalized = True

    def __str__(self):
        return "ahocorapy KeywordTree"

    def search_lss_for_children(self, zero_state):
        processed = set()
        to_process = [zero_state]
        while to_process:
            state = to_process.pop()
            processed.add(state.identifier)
            for child in state.transitions.values():
                if child.identifier not in processed:
                    self.search_lss(child)
                    to_process.append(child)

    def search_lss(self, state):
        if state.longest_strict_suffix is None:
            parent = state.parent
            if parent.longest_strict_suffix is None:
                # Has not been done yet. Do early
                self.search_lss(parent)
            traversed = parent.longest_strict_suffix
            while True:
                if state.symbol in traversed.transitions and\
                        traversed.transitions[state.symbol] != state:
                    state.longest_strict_suffix =\
                        traversed.transitions[state.symbol]
                    break
                elif traversed == self._zero_state:
                    state.longest_strict_suffix = self._zero_state
                    break
                else:
                    if traversed.longest_strict_suffix is None:
                        self.search_lss(traversed)
                    traversed = traversed.longest_strict_suffix

            for symbol, next_state in\
                    state.longest_strict_suffix.transitions.items():
                if next_state != self._zero_state\
                        and (symbol not in state.transitions or
                             next_state.success):
                    state.transitions[symbol] = next_state



from builtins import str
from builtins import object
import pygraphviz as pgv


class Visualizer(object):

    def __init__(self):
        self._added = set()

    def draw(self, filename, kwtree):
        g = pgv.AGraph(directed=True)
        for state in kwtree._states:
            if state['success']:
                g.add_node(
                    state['id'], color='green',
                    label=str(state['id']) + ' [' +
                    state['matched_keyword'] + ']')
            else:
                g.add_node(state['id'])
        for state in kwtree._states:
            if 'transitions' in state:
                for symbol_id, state_id in enumerate(state['transitions']):
                    if state_id >= 0:
                        g.add_edge(state['id'], state_id,
                                   label=kwtree._symbol_list[symbol_id])
        g.draw(filename, prog='dot')



#!/usr/bin/python
# -*- coding: utf-8 -*-

from os import path
import os

here = os.path.abspath(os.path.dirname(__file__))
try:
    import pypandoc
    long_description = pypandoc.convert_file('README.md', 'rst')
except:
    with open(path.join(here, 'README.md')) as f:
        long_description = f.read()

from setuptools import setup, find_packages



with open('%s/VERSION' % here) as f:
    __version__ = f.readline().strip()

setup(
    author='abusix',
    author_email='fp@abusix.com',
    description='ahocorapy - Pure python ahocorasick implementation',
    long_description=long_description,
    name='ahocorapy',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    install_requires=['future'],
    python_requires='>=2.7',
    url='https://github.com/abusix/ahocorapy',
    project_urls={
        'Source': 'https://github.com/abusix/ahocorapy',
        'Company': 'https://www.abusix.com/'
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords = ['keyword', 'search', 'purepython', 'aho-corasick', 'ahocorasick', 'abusix'],
    license='MIT',
    version=__version__
)
